/*
 * Copyright Elasticsearch B.V. and/or licensed to Elasticsearch B.V. under one
 * or more contributor license agreements. Licensed under the Elastic License
 * 2.0; you may not use this file except in compliance with the Elastic License
 * 2.0.
 */

import type { FunctionComponent } from 'react';
import React, { memo, useEffect, useState } from 'react';
import type { AppMountParameters } from '@kbn/core/public';
import { EuiCode, EuiEmptyPrompt, EuiErrorBoundary, EuiPanel, EuiPortal } from '@elastic/eui';
import type { History } from 'history';
import { Redirect, useRouteMatch } from 'react-router-dom';
import { Router, Routes, Route } from '@kbn/shared-ux-router';
import { FormattedMessage } from '@kbn/i18n-react';
import { i18n } from '@kbn/i18n';
import styled from 'styled-components';
import useObservable from 'react-use/lib/useObservable';
import { QueryClient, QueryClientProvider } from '@tanstack/react-query';
import { ReactQueryDevtools } from '@tanstack/react-query-devtools';

import type { TopNavMenuData } from '@kbn/navigation-plugin/public';

import { KibanaThemeProvider } from '@kbn/kibana-react-plugin/public';

import { KibanaContextProvider } from '@kbn/kibana-react-plugin/public';
import { RedirectAppLinks } from '@kbn/shared-ux-link-redirect-app';
import { EuiThemeProvider } from '@kbn/kibana-react-plugin/common';

import type { FleetConfigType, FleetStartServices } from '../../plugin';

import { PackageInstallProvider } from '../integrations/hooks';

import { type FleetStatusProviderProps, useAuthz, useFleetStatus, useFlyoutContext } from './hooks';

import {
  ConfigContext,
  FleetStatusProvider,
  KibanaVersionContext,
  sendGetPermissionsCheck,
  sendSetup,
  useBreadcrumbs,
  useStartServices,
  UIExtensionsContext,
  FlyoutContextProvider,
} from './hooks';
import {
  Error,
  Loading,
  FleetSetupLoading,
  AgentEnrollmentFlyout,
  FleetServerFlyout,
} from './components';
import type { UIExtensionsStorage } from './types';

import { FLEET_ROUTING_PATHS } from './constants';
import { DefaultLayout, DefaultPageTitle, WithoutHeaderLayout, WithHeaderLayout } from './layouts';
import { AgentPolicyApp } from './sections/agent_policy';
import { DataStreamApp } from './sections/data_stream';
import { AgentsApp } from './sections/agents';
import { MissingESRequirementsPage } from './sections/agents/agent_requirements_page';
import { CreatePackagePolicyPage } from './sections/agent_policy/create_package_policy_page';
import { EnrollmentTokenListPage } from './sections/agents/enrollment_token_list_page';
import { UninstallTokenListPage } from './sections/agents/uninstall_token_list_page';
import { SettingsApp } from './sections/settings';
import { DebugPage } from './sections/debug';
import { ExperimentalFeaturesService } from './services';

const FEEDBACK_URL = 'https://ela.st/fleet-feedback';

const queryClient = new QueryClient();

const ErrorLayout: FunctionComponent<{ isAddIntegrationsPath: boolean }> = ({
  isAddIntegrationsPath,
  children,
}) => (
  <EuiErrorBoundary>
    {isAddIntegrationsPath ? (
      <WithHeaderLayout leftColumn={<DefaultPageTitle />}>{children}</WithHeaderLayout>
    ) : (
      <DefaultLayout>
        <WithoutHeaderLayout>{children}</WithoutHeaderLayout>
      </DefaultLayout>
    )}
  </EuiErrorBoundary>
);

const Panel = styled(EuiPanel)`
  max-width: 500px;
  margin-right: auto;
  margin-left: auto;
`;

const PermissionsError: React.FunctionComponent<{ error: string }> = memo(({ error }) => {
  if (error === 'MISSING_SECURITY') {
    return <MissingESRequirementsPage missingRequirements={['security_required', 'api_keys']} />;
  }

  if (error === 'MISSING_PRIVILEGES') {
    return (
      <Panel data-test-subj="missingPrivilegesPrompt">
        <EuiEmptyPrompt
          iconType="securityApp"
          title={
            <h2 data-test-subj="missingPrivilegesPromptTitle">
              <FormattedMessage
                id="xpack.fleet.permissionDeniedErrorTitle"
                defaultMessage="Permission denied"
              />
            </h2>
          }
          body={
            <p data-test-subj="missingPrivilegesPromptMessage">
              <FormattedMessage
                id="xpack.fleet.permissionDeniedErrorMessage"
                defaultMessage="You are not authorized to access Fleet. It requires the {roleName1} Kibana privilege for Fleet, and the {roleName2} or {roleName1} privilege for Integrations."
                values={{
                  roleName1: <EuiCode>&quot;All&quot;</EuiCode>,
                  roleName2: <EuiCode>&quot;Read&quot;</EuiCode>,
                }}
              />
            </p>
          }
        />
      </Panel>
    );
  }

  return (
    <Error
      title={
        <FormattedMessage
          id="xpack.fleet.permissionsRequestErrorMessageTitle"
          defaultMessage="Unable to check permissions"
        />
      }
      error={i18n.translate('xpack.fleet.permissionsRequestErrorMessageDescription', {
        defaultMessage: 'There was a problem checking Fleet permissions',
      })}
    />
  );
});

export const WithPermissionsAndSetup: React.FC = memo(({ children }) => {
  useBreadcrumbs('base');
  const core = useStartServices();
  const { notifications } = core;

  const hasFleetAllPrivileges = useAuthz().fleet.all;

  const [isPermissionsLoading, setIsPermissionsLoading] = useState<boolean>(false);
  const [permissionsError, setPermissionsError] = useState<string>();
  const [isInitialized, setIsInitialized] = useState(false);
  const [initializationError, setInitializationError] = useState<Error | null>(null);

  const isAddIntegrationsPath = !!useRouteMatch(FLEET_ROUTING_PATHS.add_integration_to_policy);
  const isDebugPath = !!useRouteMatch(FLEET_ROUTING_PATHS.debug);

  useEffect(() => {
    (async () => {
      setIsPermissionsLoading(false);
      setPermissionsError(undefined);
      setIsInitialized(false);
      setInitializationError(null);
      try {
        setIsPermissionsLoading(true);
        const permissionsResponse = await sendGetPermissionsCheck();
        setIsPermissionsLoading(false);
        if (permissionsResponse.data?.success) {
          try {
            const setupResponse = await sendSetup();
            if (setupResponse.error) {
              setInitializationError(setupResponse.error);
            }
            if (setupResponse.data?.nonFatalErrors?.length) {
              notifications.toasts.addError(setupResponse.data.nonFatalErrors[0], {
                title: i18n.translate('xpack.fleet.setup.uiPreconfigurationErrorTitle', {
                  defaultMessage: 'Configuration error',
                }),
              });
            }
            if (!hasFleetAllPrivileges) {
              setPermissionsError('MISSING_PRIVILEGES');
            }
          } catch (err) {
            setInitializationError(err);
          }
          setIsInitialized(true);
        } else {
          setPermissionsError(permissionsResponse.data?.error || 'REQUEST_ERROR');
        }
      } catch (err) {
        setPermissionsError('REQUEST_ERROR');
      }
    })();
  }, [notifications.toasts, hasFleetAllPrivileges]);

  if (isPermissionsLoading || permissionsError) {
    return (
      <ErrorLayout isAddIntegrationsPath={isAddIntegrationsPath}>
        {isPermissionsLoading ? <Loading /> : <PermissionsError error={permissionsError!} />}
      </ErrorLayout>
    );
  }
  // Debug page moved outside of initialization to allow debugging when setup failed
  if (isDebugPath) {
    return <DebugPage setupError={initializationError} isInitialized={isInitialized} />;
  }

  if (!isInitialized || initializationError) {
    return (
      <ErrorLayout isAddIntegrationsPath={isAddIntegrationsPath}>
        {initializationError ? (
          <Error
            title={
              <FormattedMessage
                id="xpack.fleet.initializationErrorMessageTitle"
                defaultMessage="Unable to initialize Fleet"
              />
            }
            error={initializationError}
          />
        ) : (
          <FleetSetupLoading />
        )}
      </ErrorLayout>
    );
  }

  return <>{children}</>;
});

/**
 * Fleet Application context all the way down to the Router, but with no permissions or setup checks
 * and no routes defined
 */
export const FleetAppContext: React.FC<{
  startServices: FleetStartServices;
  config: FleetConfigType;
  history: AppMountParameters['history'];
  kibanaVersion: string;
  extensions: UIExtensionsStorage;
  theme$: AppMountParameters['theme$'];
  /** For testing purposes only */
  routerHistory?: History<any>;
  fleetStatus?: FleetStatusProviderProps;
}> = memo(
  ({
    children,
    startServices,
    config,
    history,
    kibanaVersion,
    extensions,
    routerHistory,
    theme$,
    fleetStatus,
  }) => {
    const darkModeObservable = useObservable(theme$);
    const isDarkMode = darkModeObservable && darkModeObservable.darkMode;

    return (
      <RedirectAppLinks
        coreStart={{
          application: startServices.application,
        }}
      >
        <startServices.i18n.Context>
          <KibanaContextProvider services={{ ...startServices, theme: { theme$ } }}>
            <EuiErrorBoundary>
              <ConfigContext.Provider value={config}>
                <KibanaVersionContext.Provider value={kibanaVersion}>
                  <KibanaThemeProvider theme$={theme$}>
                    <EuiThemeProvider darkMode={isDarkMode}>
                      <QueryClientProvider client={queryClient}>
                        <ReactQueryDevtools initialIsOpen={false} />
                        <UIExtensionsContext.Provider value={extensions}>
                          <FleetStatusProvider defaultFleetStatus={fleetStatus}>
                            <Router history={history}>
                              <PackageInstallProvider
                                notifications={startServices.notifications}
                                theme$={theme$}
                              >
                                <FlyoutContextProvider>{children}</FlyoutContextProvider>
                              </PackageInstallProvider>
                            </Router>
                          </FleetStatusProvider>
                        </UIExtensionsContext.Provider>
                      </QueryClientProvider>
                    </EuiThemeProvider>
                  </KibanaThemeProvider>
                </KibanaVersionContext.Provider>
              </ConfigContext.Provider>
            </EuiErrorBoundary>
          </KibanaContextProvider>
        </startServices.i18n.Context>
      </RedirectAppLinks>
    );
  }
);

const FleetTopNav = memo(
  ({ setHeaderActionMenu }: { setHeaderActionMenu: AppMountParameters['setHeaderActionMenu'] }) => {
    const services = useStartServices();

    const { TopNavMenu } = services.navigation.ui;

    const topNavConfig: TopNavMenuData[] = [
      {
        label: i18n.translate('xpack.fleet.appNavigation.sendFeedbackButton', {
          defaultMessage: 'Send feedback',
        }),
        iconType: 'popout',
        run: () => window.open(FEEDBACK_URL),
      },
    ];
    return (
      <TopNavMenu
        appName={i18n.translate('xpack.fleet.appTitle', { defaultMessage: 'Fleet' })}
        config={topNavConfig}
        setMenuMountPoint={setHeaderActionMenu}
      />
    );
  }
);

export const AppRoutes = memo(
  ({ setHeaderActionMenu }: { setHeaderActionMenu: AppMountParameters['setHeaderActionMenu'] }) => {
    const flyoutContext = useFlyoutContext();
    const fleetStatus = useFleetStatus();
    const { agentTamperProtectionEnabled } = ExperimentalFeaturesService.get();

    return (
      <>
        <FleetTopNav setHeaderActionMenu={setHeaderActionMenu} />

        <Routes>
          <Route path={FLEET_ROUTING_PATHS.agents}>
            <AgentsApp />
          </Route>
          <Route path={FLEET_ROUTING_PATHS.policies}>
            <AgentPolicyApp />
          </Route>
          <Route path={FLEET_ROUTING_PATHS.enrollment_tokens}>
            <EnrollmentTokenListPage />
          </Route>
          {agentTamperProtectionEnabled && (
            <Route path={FLEET_ROUTING_PATHS.uninstall_tokens}>
              <UninstallTokenListPage />
            </Route>
          )}
          <Route path={FLEET_ROUTING_PATHS.data_streams}>
            <DataStreamApp />
          </Route>

          <Route path={FLEET_ROUTING_PATHS.settings}>
            <SettingsApp />
          </Route>

          {/* TODO: Move this route to the Integrations app */}
          <Route path={FLEET_ROUTING_PATHS.add_integration_to_policy}>
            <CreatePackagePolicyPage />
          </Route>

          <Route
            render={({ location }) => {
              // BWC < 7.15 Fleet was using a hash router: redirect old routes using hash
              const shouldRedirectHash = location.pathname === '' && location.hash.length > 0;
              if (!shouldRedirectHash) {
                return <Redirect to={FLEET_ROUTING_PATHS.agents} />;
              }
              const pathname = location.hash.replace(/^#(\/fleet)?/, '');

              return (
                <Redirect
                  to={{
                    ...location,
                    pathname,
                    hash: undefined,
                  }}
                />
              );
            }}
          />
        </Routes>

        {flyoutContext.isEnrollmentFlyoutOpen && (
          <EuiPortal>
            <AgentEnrollmentFlyout
              defaultMode={
                fleetStatus.isReady && !fleetStatus.missingRequirements?.includes('fleet_server')
                  ? 'managed'
                  : 'standalone'
              }
              isIntegrationFlow={true}
              onClose={() => flyoutContext.closeEnrollmentFlyout()}
            />
          </EuiPortal>
        )}

        {flyoutContext.isFleetServerFlyoutOpen && (
          <EuiPortal>
            <FleetServerFlyout onClose={() => flyoutContext.closeFleetServerFlyout()} />
          </EuiPortal>
        )}
      </>
    );
  }
);
